using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using midianalyser;
using System.IO;
using System.Diagnostics;

namespace ASMidiConverter
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        private void AddFiles(string[] files)
        {
            foreach (string file in files)
                listView1.Items.Add(new ListViewItem(new string[] { file, "100%" }));
            if (files.Length == 1 && listView1.SelectedItems.Count == 0)
                listView1.Items[listView1.Items.Count - 1].Selected = true;
        }

        private void button1_Click(object sender, EventArgs e)
        {
            if (openFileDialog1.ShowDialog() == DialogResult.OK)
                AddFiles(openFileDialog1.FileNames);
        }

        private void button2_Click(object sender, EventArgs e)
        {
            if(folderBrowserDialog1.ShowDialog() == DialogResult.OK)
                AddFiles(Directory.GetFiles(folderBrowserDialog1.SelectedPath, "*.mid"));
        }

        private void listView1_SelectedIndexChanged(object sender, EventArgs e)
        {
            remove.Enabled = volumeBar.Enabled = resetVolume.Enabled = listView1.SelectedItems.Count > 0;
            button3.Enabled = listView1.SelectedItems.Count == 1;
        }

        private void remove_Click(object sender, EventArgs e)
        {
            foreach (ListViewItem item in listView1.SelectedItems)
                listView1.Items.Remove(item);
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            Settings1.Default.addfilefolder = openFileDialog1.InitialDirectory;
            Settings1.Default.addfilename = openFileDialog1.FileName;
            Settings1.Default.addfolderfolder = folderBrowserDialog1.SelectedPath;
            Settings1.Default.outputfolder = folderBrowserDialog2.SelectedPath;
            Settings1.Default.Save();
        }

        private void button6_Click(object sender, EventArgs e)
        {
            if (folderBrowserDialog2.ShowDialog() == DialogResult.OK)
            {
                folderBrowserDialog2.SelectedPath += folderBrowserDialog2.SelectedPath[folderBrowserDialog2.SelectedPath.Length - 1] == '\\' ? "" : "\\";
                toolStripStatusLabel2.Text = folderBrowserDialog2.SelectedPath;
            }
        }

        private void volumeBar_Scroll(object sender, EventArgs e)
        {
            foreach (ListViewItem item in listView1.SelectedItems)
                item.SubItems[1].Text = "" + volumeBar.Value + "%";
        }

        private void resetVolume_Click(object sender, EventArgs e)
        {
            foreach (ListViewItem item in listView1.SelectedItems)
                item.SubItems[1].Text = "100%";
        }

        private void toolStripStatusLabel2_Click(object sender, EventArgs e)
        {
            Process.Start("explorer.exe", toolStripStatusLabel2.Text);
        }

        bool cancel = false;
        private delegate void FileConvertedHandler(string filename);

        private void button4_Click(object sender, EventArgs e)
        {
            if (backgroundWorker1.IsBusy)
            {
                cancel = true;
                while (backgroundWorker1.IsBusy) ;
                cancel = false;
                backgroundWorker1_RunWorkerCompleted(null, null);
            }
            else
            {
                if (folderBrowserDialog2.SelectedPath == "")
                {
                    MessageBox.Show("Select Outputfolder!", "Error");
                }
                if (listView1.Items.Count == 0)
                    return;
                button4.Text = "Cancel";
                toolStripProgressBar1.Maximum = listView1.Items.Count;
                toolStripProgressBar1.Value = 0;
                ConverterData[] data = new ConverterData[listView1.Items.Count];
                int i = 0;
                foreach(ListViewItem item in listView1.Items)
                    data[i++] = new ConverterData(item.SubItems[0].Text, byte.Parse(item.SubItems[1].Text.Replace("%","")) / 100.0);
                backgroundWorker1.RunWorkerAsync(new ConverterDataPack(folderBrowserDialog2.SelectedPath, data));
            }
        }

        private void FileConverted(string file)
        {
            foreach(ListViewItem item in listView1.Items)
                if (item.Text == file)
                {
                    item.Remove();
                    break;
                }
            toolStripProgressBar1.PerformStep();
        }

        private void backgroundWorker1_DoWork(object sender, DoWorkEventArgs e)
        {
            ConverterDataPack data = (ConverterDataPack)e.Argument;
            foreach (ConverterData args in data.data)
            {
            Retry:
                if (cancel)
                    break;
                try
                {
                    MIDI midi = new MIDI(args.filename);
                    byte[] buffer = midi.Convert(args.volume);
                    FileInfo fi = new FileInfo(args.filename);
                    BinaryWriter br = new BinaryWriter(File.Open(data.filename + fi.Name.Remove(fi.Name.LastIndexOf('.')) + ".esm", FileMode.Create, FileAccess.Write));
                    br.Write(buffer);
                    br.Close();
                    this.Invoke(new FileConvertedHandler(FileConverted), new object[] { args.filename });
                }
                catch (Exception ex)
                {
                    DialogResult res = MessageBox.Show(args.filename + " could not be converted! \n" + ex.Message, "Error", MessageBoxButtons.AbortRetryIgnore);
                    switch (res)
                    {
                        case DialogResult.Abort:
                            return;
                        case DialogResult.Retry:
                            goto Retry;
                        case DialogResult.Ignore:
                            break;
                    }
                }
            }
        }

        private void backgroundWorker1_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            button4.Text = "Convert";
        }

        private void button3_Click(object sender, EventArgs e)
        {
            string filename = "";
            try
            {
                filename = listView1.SelectedItems[0].SubItems[0].Text;
                MIDI midi = new MIDI(filename);
                byte[] buffer = midi.Convert(byte.Parse(listView1.SelectedItems[0].SubItems[1].Text.Replace("%", "")) / 100.0);
                FileInfo fi = new FileInfo(filename);
                string name = fi.Name.Remove(fi.Name.LastIndexOf('.'));
                string newname = "";
                for (int i = 0; i < name.Length; i++)
                    newname += char.IsLetterOrDigit(name, i) ? name[i] : '_';
                if (char.IsDigit(newname, 0))
                    newname = '_' + newname;
                string code = "const uint8_t " + newname + "[] PROGMEM = {\r\n";
                for (int i = 0; i < buffer.Length; i++)
                {
                    code += "0x" + buffer[i].ToString("X");
                    if (i != buffer.Length - 1)
                        code += ", ";
                    if (i % 16 == 15)
                        code += "\r\n";
                }
                code += "};";
                Clipboard.SetText(code);
                MessageBox.Show("The Code was copied to Clipboard.");   
            }
            catch (Exception ex)
            {
                MessageBox.Show(filename + " could not be converted! \n" + ex.Message, "Error");
            }
        }
    }

    struct ConverterDataPack
    {
        public ConverterDataPack(string filename, ConverterData[] data)
        {
            this.filename = filename;
            this.data = data;
        }
        public string filename;
        public ConverterData[] data;
    }

    struct ConverterData
    {
        public string filename;
        public double volume;
        public ConverterData(string filename, double volume)
        {
            this.filename = filename;
            this.volume = volume;
        }
    }
}