using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Drawing;

namespace midianalyser
{
    class MIDI
    {
        byte[] data;
        int data_pos;
        const string head = "MThd";
        const string track = "MTrk";
        Sequenzer sq;
        public MIDI(string filename)
        {
            sq = new Sequenzer();
            using (FileStream file = File.Open(filename, FileMode.Open, FileAccess.Read))
            {
                data = new byte[file.Length];
                file.Read(data, 0, data.Length);
                file.Close();
            }

            data_pos = 0;

            //Open Midi File
            if (!IsNext(head))
                throw new Exception("Start enhlt nicht " + head + "! Keine Mididatei!");
            uint len = ReadInt();
            //Console.WriteLine("Lnge des Headers: " + len);
            uint format = ReadWord();
            //Console.WriteLine("Format: " + format);
            uint MTrkTracks = ReadWord();
            //Console.WriteLine("MTrkTracks: " + MTrkTracks);
            ushort time_division = ReadWord();
            if ((time_division & 0x8000) != 0)
                throw new Exception("not supported");
            //Console.WriteLine("Basisgeschwindigkeit einer Viertelnote: " + time_division);
            //Console.ForegroundColor = ConsoleColor.White;

            List<double[]> tempo_change = new List<double[]>();
            //MTrkTracks
            int x = 0;
        NEXT_TRACK:
            if (x++ < MTrkTracks && data_pos < data.Length)
            {
                sq.ResetPos();
                byte[] time = new byte[4];
                if (!IsNext(track))
                    throw new Exception("Track beginnt nicht mit " + track + "!");
                uint tracklen = ReadInt();
                //Console.BackgroundColor = ConsoleColor.DarkCyan;
                //Console.Write("Tracklnge: " + tracklen);
                //Console.BackgroundColor = ConsoleColor.Black;
                //Console.WriteLine();
                double delta_ticks = 0;
                byte current_sound = 0;
                byte last_instruction = 0;
                int tempo_pos = 0;
                double current_tempo = CalcTempo(500000, time_division);
                while (true)
                {
                    //Deltaticks
                    uint delta_add_temp = 0;
                    byte temp;
                    do
                    {
                        temp = ReadByte();
                        delta_add_temp <<= 7;
                        delta_add_temp |= (byte)(temp & 127);
                    }
                    while ((temp & 128) != 0);
                    double delta_add = delta_add_temp;
                    double add = delta_add * current_tempo;
                    while (tempo_change.Count > tempo_pos && delta_ticks + add >= tempo_change[tempo_pos][0])
                    {
                        double dif = tempo_change[tempo_pos][0] - delta_ticks;
                        delta_add -= dif / current_tempo;
                        delta_ticks = tempo_change[tempo_pos][0];
                        current_tempo = tempo_change[tempo_pos++][1];
                        add = delta_add * current_tempo;
                    }
                    delta_ticks += add;

                    temp = ReadByte();
                    if ((temp & 128) == 0)
                    {
                        data_pos--;
                        temp = last_instruction;
                    }
                    else
                        last_instruction = temp;
                    if (temp == 0xFF)
                    {
                        byte meta = ReadByte();
                        switch (meta)
                        {
                            case 0x01:
                            case 0x02:
                            case 0x03:
                            case 0x04:
                            case 0x05:
                            case 0x06:
                            case 0x07:
                                byte slen = ReadByte();
                                //Console.BackgroundColor = ConsoleColor.DarkGreen;
                                for (int i = 0; i < slen; i++)
                                    /*Console.Write((char)*/ReadByte()/*)*/;
                                //Console.WriteLine();
                                //Console.BackgroundColor = ConsoleColor.Black;
                                break;
                            case 0x20:
                                data_pos += 2;
                                break;
                            case 0x21:
                                data_pos += 2;
                                break;
                            case 0x2F:
                                if (ReadByte() == 0)
                                    goto NEXT_TRACK;
                                else
                                    throw new Exception("0x00 expected");
                            case 0x58://Time Signature
                                ReadByte();
                                for (int i = 0; i < 4; i++)
                                    time[i] = ReadByte();
                                //Console.BackgroundColor = ConsoleColor.DarkYellow;
                                //Console.WriteLine("" + time[0] + "/" + (1 << time[1]) + ", " + time[2] + ", " + time[3]);
                                //Console.BackgroundColor = ConsoleColor.Black;
                                break;
                            case 0x54:  //SMPTE Offset
                                data_pos += 6;
                                break;
                            case 0x59://Key Signature
                                data_pos += 3;
                                break;
                            case 0x51://Tempowechsel
                                ReadByte();
                                uint read = ReadInt3();
                                //Console.WriteLine("Tempo: " + 60000000 / read);
                                current_tempo = CalcTempo(read, time_division);
                                tempo_change.Add(new double[] { delta_ticks, current_tempo });
                                break;
                            default:
                                Error("Unknow Command: " + meta.ToString("X"), true);
                                break;
                        }
                    }
                    else
                    {
                        switch (temp & 0xF0)
                        {
                            case 0xA0:
                                break;
                            case 0x90:
                                byte d1 = ReadByte();
                                byte d2 = ReadByte();
                                if (d2 == 0)
                                {
                                    sq.NoteOff(d1, d2, (byte)(temp & 0xF), current_sound, (long)delta_ticks);
                                }
                                else
                                {
                                    sq.NoteOn(d1, d2, (byte)(temp & 0xF), current_sound, (long)delta_ticks);
                                }
                                break;
                            case 0x80:
                                d1 = ReadByte();
                                sq.NoteOff(d1, ReadByte(), (byte)(temp & 0xF), current_sound, (long)delta_ticks);
                                break;
                            case 0xE0: //Pitch Wheel ignorieren
                                data_pos += 2;
                                break;
                            case 0xD0://Channel pressure
                                data_pos++;
                                break;
                            case 0xC0: //Intrumenten auswahl
                                current_sound = ReadByte();
                                break;
                            case 0xB0:
                                byte controller = (byte)(ReadByte() & 0xFF);
                                switch (controller)
                                {
                                    case 7://Volume ignorieren
                                        ReadByte();
                                        break;
                                    case 10: //Stereopanorama ignorieren
                                        ReadByte();
                                        break;
                                    default:
                                        Error("Unknow Controller: " + controller + " | Data: " + ReadByte().ToString("X"), false);
                                        break;
                                }
                                break;
                            default:
                                Error("Unknow Command: " + temp.ToString("X") + "| Data: " + ReadByte().ToString("X"), true);
                                break;
                        }
                    }
                }
            }
        }

        public byte[] Convert()
        {
            return sq.Convert(1);
        }

        public byte[] Convert(double volume)
        {
            return sq.Convert(volume);
        }

        double CalcTempo(uint tempo, int timedivision)
        {
            return tempo / 1000.0 / timedivision;
        }

        void Error(string s, bool wait)
        {
            /*Console.BackgroundColor = ConsoleColor.Red;
            Console.WriteLine("Position: " + data_pos);
            Console.WriteLine(s);
            Console.BackgroundColor = ConsoleColor.Black;
            if (wait)
               Console.ReadLine();*/
        }

        byte ReadByte()
        {
            return data[data_pos++];
        }

        ushort ReadWord()
        {
            ushort ret = 0;
            for (int i = 0; i < 2; i++)
                ret = (ushort)((ret << 8) + ReadByte());
            return ret;
        }

        uint ReadInt()
        {
            uint ret = 0;
            for (int i = 0; i < 4; i++)
                ret = (uint)((ret << 8) + ReadByte());
            return ret;
        }

        uint ReadInt3()
        {
            uint ret = 0;
            for (int i = 0; i < 3; i++)
                ret = (uint)((ret << 8) + ReadByte());
            return ret;
        }

        bool IsNext(string s)
        {
            for (int i = 0; i < s.Length; i++)
                if (s[i] != (char)ReadByte())
                    return false;

            return true;
        }
    }

    class Sequenzer
    {
        List<Note> notes;
        List<Note> begin;
        long pos;
        const int viertel = 32;
        public Sequenzer()
        {
            notes = new List<Note>();
            begin = new List<Note>();
        }
        public void NoteOn(byte note, byte dyn, byte channel, byte sound, long delta_ticks)
        {
            //Console.WriteLine("New Tone: " + note+ " / " + dyn);
            pos = delta_ticks;
            if (channel == 9) //Percussion
                sound = 255;
            begin.Add(new Note(note, pos, dyn, sound));
        }
        public void NoteOff(byte note, byte dyn, byte channel, byte sound, long delta_ticks)
        {
            if (channel == 9)
                sound = 255;
            pos = delta_ticks;
            for (int i = begin.Count - 1; i >= 0; i--)
            {
                if (begin[i].note == note && begin[i].sound == sound)
                {
                    begin[i].len = (delta_ticks - begin[i].pos);
                    //Console.WriteLine("Toneoff: " + note + " | dyn: " + dyn + " | len: " + begin[i].len);
                    notes.Add(begin[i]);
                    begin.RemoveAt(i);
                    break;
                }
            }
        }
        public void ResetPos()
        {
            pos = 0;
        }
        public const int END = 0x80;
        public const int VOLUME = 0x81;
        public const int DELAY = 0x82;
        public const int SOUND = 0x83;
        public byte[] Convert(double volume)
        {
            if (notes.Count == 0)
                return new byte[0];
            notes.Sort();
            List<List<Note>> channels = new List<List<Note>>();
            channels.Add(new List<Note>());
            channels.Add(new List<Note>());
            channels[0].Add(notes[0]);
            for (int i = 1; i < notes.Count; i++)
            {
                for (int x = 0; x < channels.Count; x++)
                    if (NoteInsertable(channels[x], notes[i]))
                    {
                        notes[i].channel = (byte)x;
                        channels[x].Add(notes[i]);
                        if (x == channels.Count - 1)
                            channels.Add(new List<Note>());
                        break;
                    }
            }
            if (channels[channels.Count - 1].Count == 0)
                channels.RemoveAt(channels.Count - 1);
            //CreateBitmap(channels);
            List<Note> conotes = new List<Note>();
            for (int x = 0; x < channels.Count; x++)
            {
                long time = 0;
                for (int y = 0; y < channels[x].Count; y++)
                {
                    long delay = channels[x][y].pos - time;
                    //Delay hinzufgen
                    while (delay >= 65536)
                    {
                        conotes.Add(new Note(time, 65536, channels[x][y].channel));
                        time += 65536;
                        delay -= 65536;
                    }
                    if (delay > 0)
                        conotes.Add(new Note(time, (ushort)delay, channels[x][y].channel));
                    time = channels[x][y].end;
                }
                channels[x].Add(new Note(LastNote(channels[x]).end, (byte)x));
                conotes.AddRange(channels[x]);
            }
            conotes.Sort();
            byte[] current_sound = new byte[256];
            byte[] current_volume = new byte[256];
            long[] current_pos = new long[256];
            Array.Clear(current_pos, 0, current_pos.Length);
            Array.Clear(current_volume, 0, current_volume.Length);
            Array.Clear(current_sound, 0, current_sound.Length);
            List<byte> data = new List<byte>();
            byte channel_count = (byte)channels.Count;
            data.Add(channel_count);
            data.Add(1);
            foreach (Note note in conotes)
            {
                if (note.IsEnd)
                {
                    data.Add(END);
                }
                else
                {
                    if (note.IsDelay)
                    {
                        int delay = (int)note.len;
                        data.Add(DELAY);
                        data.Add((byte)(delay & 0xff));
                        data.Add((byte)(delay >> 8));
                    }
                    else
                    {
                        if (current_volume[note.channel] != note.dyn * 2 / channel_count)
                        {
                            data.Add(VOLUME);
                            data.Add((byte)(Math.Min(note.dyn * volume / channel_count, 255)));
                            current_volume[note.channel] = (byte)(note.dyn * 2 / channel_count);
                        }
                        if (current_sound[note.channel] != note.sound)
                        {
                            data.Add(SOUND);
                            data.Add(note.sound);
                            current_sound[note.channel] = note.sound;
                        }
                        data.Add(note.note);
                        data.Add((byte)(note.len & 0xff));
                        data.Add((byte)(note.len >> 8));
                    }
                    current_pos[note.channel] = note.end;
                }
            }

            //CreateBitmap(data.ToArray());

            return data.ToArray();
        }

        void CreateBitmap(byte[] data)
        {
            int pos = 0;
            int channels = data[pos++];
            pos++;
            bool[] active = new bool[channels];
            ushort[] duration = new ushort[channels];
            for (int i = 0; i < channels; i++)
            {
                active[i] = true;
                duration[i] = 1;
            }
            int active_channels = channels;

            List<Rectangle> rects = new List<Rectangle>();
            List<int> channelof = new List<int>();
            int time = 0;

            while (active_channels > 0)
            {
                time++;
                for (int i = 0; i < channels; i++)
                    if (active[i] && --duration[i] == 0)
                    {
                    Next:
                        byte temp = data[pos++];
                        if ((temp & 128) == 0 || temp == DELAY)
                        {
                            int delay = data[pos++];
                            delay |= data[pos++] << 8;
                            duration[i] = (ushort)delay;
                            if ((temp & 128) == 0)
                            {
                                rects.Add(new Rectangle(time, 4 * temp, delay, 4));
                                channelof.Add(i);
                            }
                        }
                        else
                        {
                            switch (temp)
                            {
                                case END:
                                    active_channels--;
                                    active[i] = false;
                                    break;
                                case VOLUME:
                                case SOUND:
                                    pos++;
                                    goto Next;
                            }
                        }
                    }
            }
            SaveBitmap(time, rects, channelof, "midiend.png");
        }

        void CreateBitmap(List<List<Note>> notes)
        {
            List<Rectangle> rects = new List<Rectangle>();
            List<int> channelof = new List<int>();
            int max = 0;
            for (int x = 0; x < notes.Count; x++)
                foreach (Note n in notes[x])
                {
                    max = Math.Max(max, (int)n.end);
                    rects.Add(new Rectangle((int)n.pos, n.note * 4, (int)n.len, 4));
                    channelof.Add(n.channel);
                }
            SaveBitmap(max, rects, channelof, "midistart.png");
        }

        void SaveBitmap(int width, List<Rectangle> rects, List<int> channelof, string filename)
        {
            Bitmap bmp = new Bitmap(width, 128 * 4);
            Graphics g = Graphics.FromImage(bmp);
            Pen[] myPens = new Pen[] { Pens.Black, Pens.Red, Pens.Green, Pens.Yellow, Pens.Blue, Pens.Orange, Pens.DarkBlue, Pens.DarkGreen, Pens.Gold, Pens.Purple, Pens.Pink };
            for (int i = 0; i < rects.Count; i++)
            {
                g.DrawRectangle(myPens[channelof[i] % myPens.Length], rects[i]);
            }
            g.Dispose();
            bmp.Save(filename);
        }

        bool NoteInsertable(List<Note> channel, Note note)
        {
            return channel.Count == 0 || LastNote(channel).end <= note.pos;
        }

        Note LastNote(List<Note> list)
        {
            return list[list.Count - 1];
        }
    }

    class Note : IComparable
    {
        public long len;
        public byte note, dyn, sound;
        public long pos;
        public byte channel = 0;
        public bool IsEnd = false;
        public bool IsDelay = false;
        public Note(long pos, uint len, byte channel)
        {
            this.pos = pos; this.len = len; this.channel = channel;
            IsDelay = true;
        }
        public Note(long pos, byte channel)
        {
            IsEnd = true;
            this.pos = pos;
            this.channel = channel;

        }
        public Note(byte note, long pos, byte dyn, byte sound)
        {
            this.note = note; this.pos = pos; this.dyn = dyn; this.sound = sound;
        }
        public override string ToString()
        {
            if (IsEnd)
                return "" + pos + " END";
            if (IsDelay)
                return "" + pos + " Delay: " + len;
            return "" + pos + " Note: " + note + " / " + len;
        }
        public int CompareTo(object note)
        {
            long comp = pos - ((Note)note).pos;
            if (comp > 0)
                return 1;
            else if (comp < 0)
                return -1;
            else
                return Math.Sign(channel - ((Note)note).channel);
        }
        public long end
        {
            get { return pos + len; }
        }
    }
}
